/*
 
 Copyright 1996 by J.D. Koftinoff Software Ltd, 
 under the KPL (Koftinoff Public License)
 
 THE NO-NONSENSE KOFTINOFF PUBLIC LICENSE (KPL)

 
 This software is provided free of charge.  You use this software only under 
 the following conditions:
 
    * You agree that you use this software at your own risk
    * You agree that the copyright of this software is held 
      by J.D. Koftinoff Software, Ltd.
 
 If you agree to these conditions, you may modify and use the software in 
 commercial as well as free projects.  However, If you use this software in a 
 commercial project, you MUST send a neat postcard to me at:
 
    Jeff Koftinoff
    J.D. Koftinoff Software, Ltd.
    #5 - 1131 Burnaby St.
    Vancouver, B.C.
    Canada
    V6E-1P3
 
    email: jeffk@jdkoftinoff.com
 
 Have Fun!
 
 P.S. Send me a postcard anyways please! I'll send one back!
 */

#include "jkcthread.h"
#include "jkthread.h"

#define DBG(a)

//---------------------------------------------------------------------------

// We rewrite the C++ memory allocation operators here to make sure 
// that they use my new thread-safe malloc and free.

void *operator new (size_t sz)
{
    return calloc(1,sz);
}

void *operator new[] (size_t sz )
{
	return calloc(1,sz);
}

void operator delete (void *p)
{
    free(p);
}

void operator delete[] (void *p)
{
	free(p);
}


//---------------------------------------------------------------------------

JKThread::JKThread()
{
	pid=0;
	done=false;
	started=false;
}

JKThread::~JKThread()
{
	if( !done && started )
	{
	    jkthread_kill( pid );	
		
		// spin until done is false. HACK alert!
		
		while( !done )
		{
		    usleep(1000); /* normally the signal gets delivered the first time around */
		}

	}	

}

bool JKThread::ThreadRun()
{
    int ret;

	// If we already started, return false
	
	if( started )
	  return false;

	// Do any inherited PreRun tasks. Return false if they fail
	
	if( !ThreadPreRun() )
	{
	    return false;	
	}
	
	// start the thread
	
	ret = jkthread_create( start, this, 0, terminate );
	
	if( ret<0 )
	{
	    return false;	
	}
	else
	{
    	return true;
	}
	
}


bool JKThread::ThreadPreRun()
{
	return true;
}


void  JKThread::ThreadKill()
{
    if( started && !done )
	{
	    jkthread_kill( pid );
		// spin until done is false. HACK alert!

		
		while( !done )
		{
		    usleep(1000);
		}

	}
}

void JKThread::PreMain()
{	
}

void JKThread::PostMain()
{	
}

void JKThread::ThreadExit()
{
    PostMain();
    done=true;
    jkthread_kill( pid );	
}

	
void JKThread::start( void *self_ )
{
	JKThread *self = (JKThread *)self_;

	// Set up the local flags
	
	self->pid = getpid();
	
	self->done=false;	
	self->started=true;
	
	// Do the PreMain() stuff
	self->PreMain();
	
	// Run the Main() thread routine
	self->Main();
	
	// Finish by running the PostMain() routine
	self->PostMain();
	
	// we are done
	self->done=true;
}

void JKThread::terminate( void *self_ )
{
	// This is called in the parent's thread when this thread 
	// is killed via a signal.
	// PostMain() is not called in this case.
	
    JKThread *self = (JKThread *)self_;
	
    // Mark as done so that the destructor (or anyone else) 
	// doesn't try to kill us again!
				  
	self->done=true;
				  
    DBG (( "\nJKThread::terminate %08lx called!\n", (long)self_ ) )
}



//---------------------------------------------------------------------------


JKThreadWithMsgQ::JKThreadWithMsgQ()
{	
	// set default size for the message Queue
	msgq_size = 64;
	msgq = 0;
}

JKThreadWithMsgQ::~JKThreadWithMsgQ()
{
	delete msgq;
}


bool JKThreadWithMsgQ::ThreadPreRun()
{
	msgq = new JKMsgQ(msgq_size);	
	return msgq!=0;
}		  


//---------------------------------------------------------------------------


JKMsgQ::JKMsgQ(int nummsgs)
    : 
    empty_sem(0),
    access_sem(1)
{
    q=new JKMsg[ nummsgs ];
	max_msgs=nummsgs;
	num_msgs=0;
	next_in=0;
	next_out=0;

}


JKMsgQ::~JKMsgQ()
{
    delete [] q;	
}

	
bool    JKMsgQ::AddMsg( 
			   unsigned long type,
			   unsigned long param1,
			   unsigned long param2,
			   void *data = 0,
			   unsigned long data_len = 0
			   )
{
	if( num_msgs<max_msgs )
	{		
        JKSemaphoreBlock    block( access_sem );
		
		int in = next_in;
		
		JKMsg *m = &q[next_in];		
				
		m->type = type;
		m->param1 = param1;
		m->param2 = param2;
		m->data = data;
		m->data_len = data_len;
		
		in++;
		if( in>=max_msgs )
		{
		    in=0;	
		}
		next_in=in;

		if( num_msgs++ == 0 )
    	    empty_sem.Release();	
		
		return true;
	}
	else
	{
	    return false;	
	}
	
}


bool    JKMsgQ::CanGetMsg() const
{
    return num_msgs>0;
}

	
bool    JKMsgQ::GetMsg( JKMsg *m)
{
    if( num_msgs==0 )
	{			
        if( !empty_sem.Get() )
		{
		    return false;	
		}		
	}
	
	
    JKSemaphoreBlock    block( access_sem );
    
	if( num_msgs==0 )
	{
		DBG(( "\n%5d: JKMsgQ::GetMsg() still empty!\n", getpid() ));
		return false;
	}
	else
	{
		int out = next_out;
		*m = q[out++];
		
		if( out>=max_msgs )
		{
		    out=0;	
		}
		
		next_out=out;
		--num_msgs;
		
        return true;
	}	
}





