/*
 
 Copyright 1996 by J.D. Koftinoff Software Ltd, 
 under the KPL (Koftinoff Public License)
 
 THE NO-NONSENSE KOFTINOFF PUBLIC LICENSE (KPL)

 
 This software is provided free of charge.  You use this software only under 
 the following conditions:
 
    * You agree that you use this software at your own risk
    * You agree that the copyright of this software is held 
      by J.D. Koftinoff Software, Ltd.
 
 If you agree to these conditions, you may modify and use the software in 
 commercial as well as free projects.  However, If you use this software in a 
 commercial project, you MUST send a neat postcard to me at:
 
    Jeff Koftinoff
    J.D. Koftinoff Software, Ltd.
    #5 - 1131 Burnaby St.
    Vancouver, B.C.
    Canada
    V6E-1P3
 
    email: jeffk@jdkoftinoff.com
 
 Have Fun!
 
 P.S. Send me a postcard anyways please! I'll send one back!
 */

#ifndef __JKTHREAD_H
#define __JKTHREAD_H

#include "jkcthread.h"

class JKThread;
    class JKThreadWithMsgQ;

class JKMsg;
class JKMsgQ;

class JKSemaphoreBase;
    class JKSemaphore;
    class JKRecursiveSemaphore;

class JKSemaphoreBlock;


//------------------------------------------------------------------------------------
class JKSemaphoreBase
{
public:
	
    virtual ~JKSemaphoreBase()
	{
	}
		
	virtual bool    Get() = 0;
	virtual bool    TryGet() = 0;	
	virtual void    Release() = 0;

};

//------------------------------------------------------------------------------------
class JKSemaphore : public JKSemaphoreBase
{
public:
	explicit JKSemaphore(int cnt=1)
	{
	    sem = jksem_create1( cnt );	
	}
	
	explicit JKSemaphore(jksem *s)
	{
	    sem = s;	
	}
	
    virtual ~JKSemaphore()
	{
	    jksem_kill( sem );	
	}
		
	bool    Get()
	{
	    return (bool)jksem_get(sem);	
	}
	
	bool    TryGet()
	{
	    return (bool)jksem_tryget(sem);
	}
	
	
	void    Release()
	{
		jksem_release(sem);
	}
	
private:
	
	jksem *sem;

};

//------------------------------------------------------------------------------------

class JKRecursiveSemaphore : public JKSemaphoreBase
{
public:
	explicit JKRecursiveSemaphore(int cnt=1)
	{
	    sem = jksem_create1( cnt );	
	}
	
	explicit JKRecursiveSemaphore(jksem *s)
	{
	    sem = s;	
	}
	
    virtual ~JKRecursiveSemaphore()
	{
	    jksem_kill( sem );	
	}
		
	bool    Get()
	{
	    return (bool)jkrsem_get(sem);	
	}
	
	bool    TryGet()
	{
	    return (bool)jkrsem_tryget(sem);
	}
	
	
	void    Release()
	{
		jkrsem_release(sem);
	}
	
private:
	
	jksem *sem;

};

//------------------------------------------------------------------------------------

class JKSemaphoreBlock
{
public:
	explicit JKSemaphoreBlock( JKSemaphoreBase *s ) : semblock( *s )
	{
		semblock.Get();
	}
	
	explicit JKSemaphoreBlock( JKSemaphoreBase &s ) : semblock( s )
	{
		semblock.Get();
	}
	
	
	virtual ~JKSemaphoreBlock()
	{
	    semblock.Release();	
	}
	
	
private:
	JKSemaphoreBase &semblock;
	
};


#define JKSEMBLOCK(a) JKSemaphoreBlock __jksemblock__##a(a)



//------------------------------------------------------------------------------------

class JKMsg
{
public:
	JKMsg()
	{
		type=0;
		param1=0;
		param2=0;
		data=0;
		data_len=0;
	}

	
	virtual ~JKMsg()
	{
	    delete data;
	}
	
	unsigned long type;
	unsigned long param1;
	unsigned long param2;
	void *data;
	unsigned long data_len;
};


//------------------------------------------------------------------------------------

class JKMsgQ
{
public:
	explicit JKMsgQ(int nummsgs);
	virtual ~JKMsgQ();

	// AddMsg returns false if there is no space. It does not block
	
	bool    AddMsg( 
				   unsigned long type,
				   unsigned long param1,
				   unsigned long param2,
				   void *data = 0,
				   unsigned long data_len = 0
				   );

	bool    CanGetMsg() const;
	
	// GetMsg() will block if there is no message available.
	
	bool    GetMsg( JKMsg *m );

 private:
	JKMsg *q;
	int max_msgs;
	int num_msgs;
	int next_in;
	int next_out;
	
	JKSemaphore empty_sem;
	JKSemaphore access_sem;	
};


//------------------------------------------------------------------------------------

class JKThread
{
public:
	 JKThread();
	virtual ~JKThread();

	// Call ThreadSetInitialStack before calling ThreadRun() 
	// if you want a non-default sized stack	
	void    ThreadSetInitialStack( int stksize_ )
	{
	    stksize = stksize_;	
	}
	
	
	// Call ThreadRun() once only to start the thread running	
	bool    ThreadRun();
	
	// Call ThreadKill() to kill this thread if it is misbehaving
	void    ThreadKill();
	
	pid_t   ThreadGetPID() const
	{
		return pid;
	}	
	
	// ThreadHasStarted() returns true only if the actual thread has
	// started executing.
	
	bool    ThreadHasStarted() const
	{
		return started;
	}
	
	// ThreadIsDone() returns true either if Main() returned or
	// the thread was killed
	
	bool    ThreadIsDone() const
	{
	    return done;	
	}
	
							  
protected:
	
	// PreMain() is called from the thread context, just before Main() is called
    virtual void PreMain();
	
	// Main() is where your main thread routine lives
	virtual void Main() = 0;
	
	// PostMain() is called after the Main() finished or when Exit() is called.
	// PostMain() is called from the thread context.
    virtual void PostMain();	
	
	// ThreadPreRun() is called from the parent's thread context, just before 
	// the thread is spawned.  If PreRun() returns false, the thread will
	// not be created.
    virtual bool ThreadPreRun();	  

    // ThreadExit() can be called from your thread to nicely abort the thread.
    // After Exit() is called, PostMain() will be called to clean up.
	virtual void ThreadExit();

private:
	pid_t pid;
	int stksize;
	volatile bool done;
	volatile bool started;
	
	static  void start( void *);	
	static  void terminate( void *);
};


//------------------------------------------------------------------------------------


class JKThreadWithMsgQ : public JKThread
{
public:
	 JKThreadWithMsgQ();
	virtual ~JKThreadWithMsgQ();

	void    ThreadSetInitialQSize( int qsize_ )
	{
	    msgq_size = qsize_;	
	}
	
	bool    ThreadPostMsg( 
				   unsigned long type,
				   unsigned long param1,
				   unsigned long param2,
				   void *data = 0,
				   unsigned long data_len = 0
				   )
	{
	    msgq->AddMsg( type, param1, param2, data, data_len );	
	}
	
protected:
	virtual bool ThreadPreRun();
	
	bool    ThreadCanGetMsg() const
	{
		return msgq->CanGetMsg();
	}
	
	bool    ThreadGetMsg( JKMsg *m )
	{
	    return msgq->GetMsg(m);	
	}	  

private:
	int msgq_size;

	JKMsgQ  *msgq;
};


//------------------------------------------------------------------------------------

#endif
